# canonical camera setting and basic data setting
# we set it same as the E300 camera (crop version)
#  
data_basic=dict(
    canonical_space = dict(
        img_size=(540, 960),
        focal_length=1196.0,
    ),
    depth_range=(0.9, 150),
    depth_normalize=(0.006, 1.001),
    crop_size = (512, 960),
    clip_depth_range=(0.9, 150),
)    
