# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
from .sevenscenes import VislocSevenScenes
from .cambridge_landmarks import VislocCambridgeLandmarks
from .aachen_day_night import VislocAachenDayNight
from .inloc import VislocInLoc
