class ProjectionRegistry:
    @classmethod
    def register(cls, name: str): ...
    @classmethod
    def create(cls, name: str, **kwargs): ...
    @classmethod
    def available_projections(cls) -> list: ...
