"""
panorai_data_rewritten.pcd
==========================

This sub-package provides point-cloud (PCD) structures, handlers,
and various blending utilities to merge or process 3D data.
"""

# Typically, import only user-facing classes/functions here:
# from .data import PCD
# from .handler import PCDHandler
# from .blender import SomeBlenderClass

# __all__ = ["PCD", "PCDHandler", ...]