class PanoraiFactoryError(Exception): ...

class SamplerNotFoundError(PanoraiFactoryError):
    def __init__(self, name, available) -> None: ...

class BlenderNotFoundError(PanoraiFactoryError):
    def __init__(self, name, available) -> None: ...

class ProjectionNotFoundError(PanoraiFactoryError):
    def __init__(self, name, available) -> None: ...

class ConfigNotFoundError(PanoraiFactoryError):
    def __init__(self, name, available) -> None: ...
