import numpy as np
from _typeshed import Incomplete

class ArrayLikeData:
    def __array__(self, dtype: Incomplete | None = None, copy: bool = True) -> np.ndarray: ...
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs): ...
    def __getitem__(self, index) -> np.ndarray | ArrayLikeData: ...
    def __setitem__(self, index, value) -> None: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
