from .registry import ConfigRegistry as ConfigRegistry
from _typeshed import Incomplete
from typing import Any

logger: Incomplete

class ConfigManager:
    @classmethod
    def create(cls, name: str, **kwargs) -> Any: ...
    @classmethod
    def get(cls, name: str) -> Any: ...
    @classmethod
    def modify_config(cls, name: str, **kwargs): ...
    @classmethod
    def get_config_parameters(cls, name: str) -> dict[str, Any]: ...
    @classmethod
    def describe_config(cls, name: str, output_format: str = 'pretty') -> None: ...
    @classmethod
    def reset(cls, name: str | None = None): ...
    @classmethod
    def get_all_configs(cls) -> dict[str, Any]: ...
