class PanoraiError(Exception): ...
class InvalidDataError(PanoraiError): ...
class ChannelMismatchError(PanoraiError): ...
class MetadataValidationError(PanoraiError): ...
class DataConversionError(PanoraiError): ...
class MissingChannelError(PanoraiError): ...
class FaceSetError(PanoraiError): ...
class ImageProcessingError(PanoraiError): ...
class ProjectionError(Exception): ...
class ConfigurationError(ProjectionError): ...
class RegistrationError(ProjectionError): ...
class ProcessingError(ProjectionError): ...
class GridGenerationError(ProjectionError): ...
class TransformationError(ProjectionError): ...
class InterpolationError(ProjectionError): ...
