import numpy as np
import open3d as o3d

class PCD:
    def __init__(self, points: np.ndarray, colors: np.ndarray, radius_image: np.ndarray = None, indexes: np.ndarray = None, shape: tuple = None) -> None: ...
    def hook_bias_adjustment(self, func): ...
    @property
    def points(self) -> np.ndarray: ...
    @property
    def indexes(self) -> np.ndarray: ...
    @property
    def shape(self) -> np.ndarray: ...
    @property
    def radius_image(self) -> np.ndarray: ...
    @property
    def colors(self) -> np.ndarray: ...
    @property
    def o3d(self) -> o3d.geometry.PointCloud: ...
